(* This file is distributed under the terms of the MIT License, also
   known as the X11 Licence.  A copy of this license is in the README
   file that accompanied the original distribution of this file.

   Based on code written by:
     Brian Aydemir *)

(** Assorted facts about lists that are not in Coq's standard library.

    Implicit arguments are declared by default in this library. *)

Set Implicit Arguments.

Require Import Coq.Lists.List.
Require Import Coq.Lists.SetoidList.

Open Scope list_scope.


(* ********************************************************************** *)
(** * List structure *)

Lemma cons_eq_app : forall (A : Type) (z : A) (xs ys zs : list A),
  z :: zs = xs ++ ys ->
  (exists qs, xs = z :: qs /\ zs = qs ++ ys) \/
  (xs = nil /\ ys = z :: zs).
Proof.
  destruct xs; intros ? ? H; simpl in *.
    auto.
    injection H; intros; subst; eauto.
Qed.

Lemma app_eq_cons : forall (A : Type) (z : A) (xs ys zs : list A),
  xs ++ ys = z :: zs ->
  (exists qs, xs = z :: qs /\ zs = qs ++ ys) \/
  (xs = nil /\ ys = z :: zs).
Proof. auto using cons_eq_app. Qed.

Lemma nil_eq_app : forall (A : Type) (xs ys : list A),
  nil = xs ++ ys ->
  xs = nil /\ ys = nil.
Proof. auto using List.app_eq_nil. Qed.

Lemma app_cons_not_nil : forall (A : Type) (y : A) (xs ys : list A),
  xs ++ y :: ys <> nil.
Proof.
  intros A y xs ys H.
  symmetry in H. revert H.
  apply List.app_cons_not_nil.
Qed.


(* ********************************************************************** *)
(** * List membership *)

Lemma In_map : forall (A B : Type) (xs : list A) (x : A) (f : A -> B),
  In x xs ->
  In (f x) (map f xs).
Proof.
  induction xs; intros ? ? H; simpl in *.
    auto.
    destruct H as [H | H]; subst; auto.
Qed.


(* ********************************************************************** *)
(** * Setoid facts *)

Lemma InA_In : forall (A : Type) (x : A) (xs : list A),
  InA (@eq _) x xs -> In x xs.
Proof.
  induction xs; intros H.
    inversion H.
    inversion H; subst; simpl in *; auto.
Qed.

Lemma InA_iff_In : forall (A : Type) (x : A) (xs : list A),
  InA (@eq _) x xs <-> In x xs.
Proof. split; auto using InA_In, SetoidList.In_InA. Qed.
